import requests

from pycamel.src.modules.core.filter import Filter
from pycamel.src.modules.response.response import CamelResponse


class Router:

    def __init__(self, path: str) -> None:
        self.path = path
        self.headers = {'Content-Type': 'application/json'}

        self.request_path = path
        self.request_headers = self.headers

        self._execution_method = None

    def get(self, *args, **kwargs) -> CamelResponse:
        """
        Request method based on :class:`Request` of requests lib.
        Gets request method as object and makes request.
        :param args: Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
        :param kwargs: Optional arguments that ``request`` takes.
        :return: Result of execution _fetch method. class CamelResponse.
        """
        self._execution_method = requests.get
        return self._fetch(*args, **kwargs)

    def post(self, *args, **kwargs) -> CamelResponse:
        """
        Request method based on :class:`Request` of requests lib.
        Gets request method as object and makes request.
        :param args: Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
        :param kwargs: Optional arguments that ``request`` takes.
        :return: Result of execution _fetch method. class CamelResponse.
        """
        self._execution_method = requests.post
        return self._fetch(*args, **kwargs)

    def put(self, *args, **kwargs) -> CamelResponse:
        """
        Request method based on :class:`Request` of requests lib.
        Gets request method as object and makes request.
        :param args: Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
        :param kwargs: Optional arguments that ``request`` takes.
        :return: Result of execution _fetch method. class CamelResponse.
        """
        self._execution_method = requests.put
        return self._fetch(*args, **kwargs)

    def patch(self, *args, **kwargs) -> CamelResponse:
        """
        Request method based on :class:`Request` of requests lib.
        Gets request method as object and makes request.
        :param args: Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
        :param kwargs: Optional arguments that ``request`` takes.
        :return: Result of execution _fetch method. class CamelResponse.
        """
        self._execution_method = requests.patch
        return self._fetch(*args, **kwargs)

    def delete(self, *args, **kwargs) -> CamelResponse:
        """
        Request method based on :class:`Request` of requests lib.
        Gets request method as object and makes request.
        :param args: Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
        :param kwargs: Optional arguments that ``request`` takes.
        :return: Result of execution _fetch method. class CamelResponse.
        """
        self._execution_method = requests.delete
        return self._fetch(*args, **kwargs)

    def add_to_path(self, parameter: str) -> 'Router':
        self.request_path += parameter
        return self

    def set_headers(self, headers):
        self.request_headers = headers
        return self

    def set_filters(self, filters: dict) -> 'Router':
        self.request_path += Filter.build_filter(filters)
        return self

    def append_header(self, header_key, header_value):
        self.request_headers[header_key] = header_value
        return self

    def _clear(self) -> None:
        self.request_path = self.path
        self.request_headers = self.headers

    def _fetch(self, *args, **kwargs) -> CamelResponse:
        response = self._execution_method(
            url=self.request_path,
            headers=self.request_headers,
            *args,
            **kwargs
        )
        self._clear()
        return CamelResponse(response=response, headers=self.request_headers)



"""
:authors: Yurii Abramenko
:license: Apache License, Version 2.0, see LICENSE file
:copyright: (c) 2022 SolveMeSolutions
"""

from .src.modules.core.config import CamelConfig
from .src.modules.core.filter import Filter
from .src.modules.core.validator import Validator

from .src.modules.response.response import CamelResponse

from .src.modules.routing.router import Router
from .src.modules.routing.router_maker import RouterMaker

from .src.utils.camel_enum import CamelEnum


__author__ = 'Yurii Abramenko'
__version__ = '0.0.2'
__email__ = 'yura.abramenko1@gmail.com'



from enum import Enum


class CamelEnum(Enum):
    """
    Default inheritance from Enum class with possibility to get list of enums.
    """
    @classmethod
    def list(cls) -> list:
        """
        Method gets all values from Enum class and prepare array with them.
        :return: list of values of Enum instance
        """
        return list(map(lambda c: c.value, cls))



import os


class CamelConfig:
    """
    Configuration class is responsible for project configuration.
    Parameters that were set for it decides how it will work.
    """
    def __init__(self, host: str, project_validation_key: str = None) -> None:
        """
        :param host: Base url for all services and endpoints.
            If we have something like that:
            https://google.com/v2/api/get_urls?is_public=true
            so, url for that property will be https://google.com/
        :param project_validation_key: It is not mandatory parameter.
            Receives string that needs for getting data from response object.
            Validation method get that parameter in case when it didn't set
            for concreate endpoint or validation function didn't received it
            directly.
            For example, if in your backend project you have stable contract
            like that:
            {"meta": {"some":"data"}, "data: {"some": "data"}}
            You don't need to get key "data" all time from response.json(),
            all that you need, just put your key here and for all endpoints we
            will try to get data by that key.
                For cases when you need to get data from lower level, you can
            set list of keys as string with :.
            Like that - "data:some:needed:"
        """
        self.host = host
        self.project_validation_key = project_validation_key
        self._set_env_properties()

    def _set_env_properties(self) -> None:
        """
        Sets all project configuration variables as env variables.
        All properties that doesn't have values, will not be set.
        :return: None
        """
        env_variables = self.__dict__
        for variable in env_variables:
            if env_variables.get(variable) is not None:
                os.environ[f"pc_{variable}"] = env_variables.get(variable)



from pycamel.src.utils.camel_enum import CamelEnum


class Statuses(CamelEnum):
    ACTIVE = "ACTIVE"
    BANNED = "BANNED"


def test_getting_list_of_attributes():
    assert isinstance(Statuses.list(), list)
    assert Statuses.list() == ['ACTIVE', 'BANNED']



import pytest

from pycamel.src.modules.core.filter import Filter


@pytest.mark.parametrize(
    "filter_value, expected_filter", [
        ({"gender": "male", "age": 18}, "?gender=male&age=18"),
        ({"page-limit": 100, "page_offset": 20},
         "?page-limit=100&page_offset=20"),
        ({"flat_id_in": [2, 34, 45, 6], "human_id_in": [411]},
         "?flat_id_in=2,34,45,6&human_id_in=411")
    ])
def test_filter_build(filter_value, expected_filter):
    generated_filter = Filter.build_filter(filter_value)
    assert generated_filter == expected_filter



from typing import List

import requests
from json import JSONDecodeError

from pydantic import BaseModel
from pydantic.error_wrappers import ValidationError
from pycamel.src.modules.core.validator import Validator


class CamelResponse:

    def __init__(
            self,
            response: requests.Response,
            headers: dict
    ) -> None:
        self.response = response
        self.headers = headers
        self.validated_objects = []
        try:
            self.response_data = response.json()
        except JSONDecodeError:
            self.response_data = {}

    def assert_status_code(
            self,
            expected_status_code: List[int]
    ) -> 'CamelResponse':
        assert self.response.status_code == expected_status_code, self
        return self

    def get_response_json(self) -> dict:
        return self.response_data

    def validate(
            self,
            schema: BaseModel,
            validation_key: str = None
    ) -> 'CamelResponse':
        try:
            self.validated_objects = Validator(
                schema, self.response_data, validation_key
            ).fetch()
            return self
        except ValidationError as validation_exception:
            raise AssertionError(
                "Received objects could not be mapped to schema"
            ) from validation_exception

    def get_validated_objects(self) -> list:
        return self.validated_objects

# TODO Refactor error log. Split request and response data.
# TODO Add to error log request body if it was sent.
    def __str__(self) -> str:
        return f"\n\nRequest sent to: {self.response.url}\n" \
               f"Response status code: {self.response.status_code}\n" \
               f"Response data: {self.response_data}\n" \
               f"Headers: {self.headers}\n" \
               f"Request time took: {self.response.elapsed}"



import os

from pycamel.src.modules.routing.router import Router


class RouterMaker:

    def __init__(self, service_host: str) -> None:
        """
        Class constructor.
        :param service_host: Receives part of url.
            For example, if main url looks like that:
            https://google.com/api-service/v1/some-endpoint
            router should have that part - api-service/v1
            After it, u can easy make any endpoints on that url
        """
        self.service_host = service_host

    def _build_url(self, route: str) -> str:
        """
        Build url from all parts of received data.
        :param route: String. Example of path /some-endpoint
        :return: String. Full path.
        """
        host = os.environ.get('pc_host')
        return f"{host}{self.service_host}{route}"

    def make_router(self, route: str) -> Router:
        """
        Returns Router object according to received path.
        Example of path /some-endpoint
        As a result u will get all REST methods with full functionality
        for that entity.
        :param route: String. Example /some-endpoint
        :return: Router object
        """
        path = self._build_url(route)
        return Router(path=path)



class Filter:
    """
    Aggregation class for functions that works with filtering and filter
    generation.
    """
    @classmethod
    def _prepare_array(cls, filter_in_items: list) -> str:
        """
        Method for deserialize array of Any to string.
        For example: List[1,2,3,4,5,] -> Str "1,2,3,4,5"
        :param filter_in_items: List of items
        :return: String row with concatenation of all array items.
        """
        if len(filter_in_items) == 1:
            return str(*filter_in_items)
        filter_result = ""
        for item in filter_in_items[:-1]:
            filter_result += f"{item},"
        filter_result += str(filter_in_items[-1])
        return filter_result

    @classmethod
    def build_filter(cls, filters: dict) -> str:
        """
        Creates url filter according to received dict.
        For example: {"age": 22, "gender": "male"} to "?age=22&gender=male"
        :param filters: Dict with filter items.
        :return: String like that "?age=22&gender=male".
        """
        final_filter_row = ""
        for item in filters.keys():
            if isinstance(filters[item], list):
                filter_item = Filter._prepare_array(filters[item])
            else:
                filter_item = filters[item]
            final_filter_row += f"&{item}={filter_item}"
        final_filter_row = final_filter_row[:0] + "?" + final_filter_row[1:]
        return final_filter_row
